unit cruStrSplitter;

interface

uses Classes;

type
  TStrSplitter = class(TObject)
  private
    fInputStr: string;
    fDelimiter: string;
    fDelLength: Integer;
    fCount: Integer;
    fParts: Array of String;
    function Get(Index: Integer): string;
    procedure CountDelimiters;
    procedure Split;
    procedure Reset;
    procedure SetPart(aIndex: Integer; aString: string);
  public
    constructor Create; overload;
    constructor Create(const aString: string; const aDelimiter: string = ';'); overload;

    procedure Execute(const aString: string; const aDelimiter: string = ';');
    procedure AssignTo(Strings: TStrings);
    procedure AppendTo(Strings: TStrings);

    property Parts[Index: Integer]: string read Get; default;
    property Count: Integer read fCount;
  end;

implementation

{ copied from StrUtils }
function PosEx(const SubStr, S: string; Offset: Integer = 1): Integer;
var
  I,X: Integer;
  Len, LenSubStr: Integer;
begin
  if Offset = 1 then
    Result := Pos(SubStr, S)
  else
  begin
    I := Offset;
    LenSubStr := Length(SubStr);
    Len := Length(S) - LenSubStr + 1;
    while I <= Len do
    begin
      if S[I] = SubStr[1] then
      begin
        X := 1;
        while (X < LenSubStr) and (S[I + X] = SubStr[X + 1]) do
          Inc(X);
        if (X = LenSubStr) then
        begin
          Result := I;
          exit;
        end;
      end;
      Inc(I);
    end;
    Result := 0;
  end;
end;

{ TStrSplitter }

procedure TStrSplitter.AppendTo(Strings: TStrings);
var
  i: Integer;
begin
  if fCount > 1 then
    for i := 0 to fCount - 1 do
      Strings.Add(fParts[i])
  else
    Strings.Add(fInputStr);
end;

procedure TStrSplitter.AssignTo(Strings: TStrings);
begin
  Strings.Clear;
  AppendTo(Strings);
end;

procedure TStrSplitter.CountDelimiters;
var
  Offset: Integer;
begin
  Offset := 1;
  repeat
    if fCount = 0 then
      Offset := PosEx(fDelimiter, fInputStr, 1)
    else
      Offset := PosEx(fDelimiter, fInputStr, Offset + fDelLength);
    Inc(fCount);
  until Offset = 0;
end;

constructor TStrSplitter.Create(const aString, aDelimiter: string);
begin
  Create;
  Execute(aString, aDelimiter);
end;

constructor TStrSplitter.Create;
begin
  inherited;
end;

function TStrSplitter.Get(Index: Integer): string;
begin
  Result := '';
  if (Index >= fCount) or (Index < 0) then Exit;

  if (fCount > 1) then
    Result := fParts[index]
  else
    Result := fInputStr;
end;

procedure TStrSplitter.Execute(const aString, aDelimiter: string);
begin
  Reset;
  fInputStr  := aString;
  fDelimiter := aDelimiter;
  fDelLength := Length(fDelimiter);
  CountDelimiters;
  if fCount > 1 then Split;
end;

procedure TStrSplitter.Reset;
begin
  fInputStr := '';
  fDelimiter := '';
  fDelLength := 0;
  fCount := 0;
  SetLength(fParts, 0);
end;

procedure TStrSplitter.SetPart(aIndex: Integer; aString: string);
begin
  fParts[aIndex] := aString;
end;

procedure TStrSplitter.Split;
var
  From, Offset, Index: Integer;
begin
  Offset := 1;
  Index := -1;
  SetLength(fParts, fCount);

  repeat
    Inc(Index);

    if Index = 0 then
      From := 1
    else
      From := Offset + fDelLength;

    Offset := PosEx(fDelimiter, fInputStr, From);

    if OffSet > 0 then
      SetPart(Index, Copy(fInputStr, From, Offset - From));
  until Offset = 0;
  SetPart(Index, Copy(fInputStr, From, Length(fInputStr) - (From - 1)));
end;

end.
